WoWUnit = {}

function WoWUnit:GetMocks()
	self.mocks = self.mocks or {};
	return self.mocks;
end

function WoWUnit:GetTestSuites()
	self.suites = self.suites or {};
	return self.suites;
end

local function copyValue(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, _copy(getmetatable(object)))
    end
    return _copy(object)
end

function WoWUnit:Mock(overrides)
	local mocks = self:GetMocks();
	
	for name, value in pairs(overrides) do
		local mock = {};
		mock.name = name;
		mock.value = _G[name];
		_G[name] = copyValue(value);
		table.insert(mocks, mock);
	end
end

function WoWUnit:RemoveMocks()
	local mocks = self:GetMocks();
	
	while #mocks > 0 do
		local mock = mocks[#mocks];
		table.remove(mocks, #mocks);
		_G[mock.name] = mock.value;
	end
end

function WoWUnit:AddTestSuite(name, suite)
	if type(name) ~= "string" then return false, "Suite name must be a string" end;
	if type(suite) ~= "table" then return false, "Suite must be a table" end;
	
	local suites = self:GetTestSuites();
	suites[name] = suite;
end

function WoWUnit:QueueTestSuite(name)
	self.queue = nil;
	self.results = nil;
	
	local suites = self:GetTestSuites();
	local suite = suites[name];
	if suite == nil then return false end;

	self:QueueSuiteHelper(suite, name);
	
	return true;
end

function WoWUnit:QueueSuiteHelper(suite, name)
	local setUp = suite.setUp;
	local tearDown = suite.tearDown;
	local mocks = suite.mocks;
	
	if type(setUp) ~= "function" then setUp = nil end;
	if type(tearDown) ~= "function" then tearDown = nil end;
	if type(mocks) ~= "table" then mocks = nil end;
	
	for k, v in pairs(suite) do
		if (k ~= "mocks") and (k ~= "setUp") and (k ~= "tearDown") then
			if (string.find(k, "test", 1) == 1) and (type(v) == "function") then
				local test = {};
				test.suite = name;
				test.name = k;
				test.setUp = setUp;
				test.tearDown = tearDown;
				test.mocks = mocks;
				test.test = v;
				WoWUnit:QueueTest(test);
			elseif type(v) == "table" then
				self:QueueSuiteHelper(v, name .. "-" .. k);
			end
		end
	end
	
	return true;
end

function WoWUnit:QueueTest(test)
	if self.queue then
		test.next = self.queue;
	end
	self.queue = test;
end

function WoWUnit:GetRemainingTestCount()
	local count = 0;
	local current = self.queue;
	while current do
		count = count + 1;
		current = current.next;
	end
	return count;
end

function WoWUnit:RunTest()
	local test = self.queue;
	if test == nil then return end;
	
	self.queue = test.next;
	
	local result = {};
	result.suite = test.suite;
	result.test = test.name;
	
	local framework = {};
	if test.setUp then framework = test.setUp(); end;
	
	if test.mocks then self:Mock(test.mocks) end;
	self:Mock(self.assertFunctions);
	
	result.pass, result.message = pcall(test.test);
	if result.pass == true then result.message = nil end;
	
	self:RemoveMocks();
	
	if test.tearDown then test.tearDown(framework) end;
	
	self:AddResult(result);
end

function WoWUnit:AddResult(result)
	if self.results then
		result.next = self.results;
	end
	self.results = result;
end

function WoWUnit:GetFirstResult()
	return self.results;
end

WoWUnit.assertFunctions = {
	
	assertEquals = function(message, a, b)
		local message = a;
		if b == nil then
			b = a;
			a = message;
			message = "";
		else
			message = message .. ": ";
		end
		
		message = message .. ("expected:<" .. (a or "nil") .. "> but was:<" .. (b or "nil") .. ">");
		if type(a) ~= type(b) then error(message) end;
		if type(a) == "table" then
			-- Compare the tables here.
		else
			if a ~= b then error(message) end;
		end
	end
	
}
